﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Threading;
using AdaTreeNat;
using Plossum.CommandLine;
using XmlClassifier.Algorithms;
using XmlClassifier.DataStructures;

namespace XmlStreamClassifier.ConsoleApplication
{
	public class Program
	{
		public static void Main(string[] args)
		{
			Options options = ParseParameters();

			List<Document> dataset = Read(options.Dataset);

			if (options.GenerateArff)
			{
				GenerateArff(options);
			}
			else
			{
				Classify(options, dataset);
			}
		}

		private static void Classify(Options options, List<Document> dataset)
		{
			long memoryCounter = 0;
			int benedictNumberBatch = 0; //Number of classified batches...:-)

			Stopwatch hammertime = new Stopwatch();

			StreamClassifier classifier = new StreamClassifier(options.Minsup, options.BatchSize, options.SynchronizedUpdate, options.Window, options.WindowSize, options.WeightSize, options.WeightSupport);

			foreach (var document in dataset)
			{
				hammertime.Start();
				classifier.ClassifyDocument(document);
				hammertime.Stop();

				classifier.UpdateModel(document);

				if (classifier.DocumentCount > 0 && classifier.DocumentCount % options.SamplingFrequency == 0)
				{
					if (options.DisplayIntermediateResults)
					{
						DisplayIntermediateResults(options, hammertime, classifier);
					}

					classifier.ClearSamplingStats();
					memoryCounter += Process.GetCurrentProcess().PrivateMemorySize64;
					benedictNumberBatch++;
				}
			}

			if (classifier.DocumentCount % options.SamplingFrequency != 0)
			{
				memoryCounter += Process.GetCurrentProcess().PrivateMemorySize64;
				benedictNumberBatch++;

				if (options.DisplayIntermediateResults)
				{
					DisplayIntermediateResults(options, hammertime, classifier);
				}
			}

			DisplayResults(options, benedictNumberBatch, memoryCounter, hammertime, classifier);
		}

		private static void DisplayResults(Options options, int numberOfBatches, long memoryCounter, Stopwatch timer, StreamClassifier classifier)
		{
			Console.Write(classifier.TotalAccuracy + "\t");

			if (options.DisplayTime)
			{
				Console.Write(timer.ElapsedMilliseconds + "\t");
				Console.Write(((double)timer.ElapsedMilliseconds / (double)numberOfBatches) + "\t");
			}

			if (options.DisplayMemory)
			{
				Console.Write(((double)memoryCounter / (double)numberOfBatches) + "\t");
			}

			Console.WriteLine();
		}

		private static void DisplayIntermediateResults(Options options, Stopwatch timer, StreamClassifier classifier)
		{
			if (options.DisplayLastBatchAccuracy)
			{
				Console.Write(classifier.LastBatchAccuracy + "\t");
			}

			Console.Write(classifier.TotalAccuracy + "\t");

			if (options.DisplayTime)
			{
				Console.Write(timer.ElapsedMilliseconds + "\t");
			}

			if (options.DisplayMemory)
			{
				Console.Write(Process.GetCurrentProcess().PrivateMemorySize64 + "\t");
			}

			Console.WriteLine();
		}

		private static List<Document> Read(string path)
		{
			List<Document> result = new List<Document>();

			using (StreamReader sr = new StreamReader(path))
			{
				while (!sr.EndOfStream)
				{
					result.Add(new Document(sr.ReadLine()));
				}
			}

			return result;
		}

		private static Options ParseParameters()
		{
			Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");

			Options options = new Options();
			CommandLineParser parser = new CommandLineParser(options, NumberFormatInfo.CurrentInfo);
			parser.Parse();
			if (!options.NoInfo)
			{
				Console.WriteLine(parser.UsageInfo.GetHeaderAsString(78));
			}

			if (options.Help)
			{
				Console.WriteLine(parser.UsageInfo.GetOptionsAsString(78));
			}
			else if (parser.HasErrors)
			{
				Console.WriteLine(parser.UsageInfo.GetErrorsAsString(78));
			}
			return options;
		}

		private static void GenerateArff(Options options)
		{
			int numOfAttrs = 20;
			string datasetName = Path.GetFileNameWithoutExtension(options.Dataset);
			List<Document> dataset = Read(options.Dataset);

			ArffGenerator generator = new ArffGenerator(options.Minsup, options.BatchSize, options.SynchronizedUpdate, options.WeightSize, options.WeightSupport);

			using (StreamWriter output = new StreamWriter(datasetName + ".arff"))
			{
				output.WriteLine("@relation " + datasetName);
				output.WriteLine();

				for (int i = 0; i < numOfAttrs; i++)
				{
					output.WriteLine("@attribute pattern" + i + " {0,1}");
				}

				output.WriteLine("@attribute class {0,1,2,3}");
				output.WriteLine();
				output.WriteLine("@data");

				foreach (var document in dataset)
				{
					output.WriteLine(generator.GetTuple(document, numOfAttrs));
				}
			}

			Console.WriteLine("Arff dataset successfully generated to: " + datasetName + ".arff");
		}
	}
}
