﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using AdaTreeNat;
using XmlClassifier.Algorithms;
using XmlClassifier.DataStructures;

namespace XmlStreamClassifier.ConsoleApplication
{
	/// <summary>
	/// XML stream classifier
	/// </summary>
	public class StreamClassifier
	{
		private List<ClassModel> models;
		private List<Document> documentBuffer;

		private double minsup;
		private bool synchronizedUpdate;
		private int treeMiningBatchSize;

		private bool isWindow;
		private int windowSize;

		private int numberOfCorrectlyAssignedDocumentsInLastBatch;
		private int numberOfDocumentsInSamplingBatch;
		private int totalNumberOfCorrectlyAssignedDocuments;
		private int totalNumberOfDocuments;

		/// <summary>
		/// Number of documents classified so far
		/// </summary>
		public int DocumentCount { get { return this.totalNumberOfDocuments; } }

		/// <summary>
		/// Classification accuracy for documents in the last batch
		/// </summary>
		public double LastBatchAccuracy { get { return (double)this.numberOfCorrectlyAssignedDocumentsInLastBatch / (double)this.numberOfDocumentsInSamplingBatch; } }

		/// <summary>
		/// Classification accuracy for all documents processed so far
		/// </summary>
		public double TotalAccuracy { get { return (double)this.totalNumberOfCorrectlyAssignedDocuments / (double)this.totalNumberOfDocuments; } }

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="minsup">Minimal support of frequent trees for a single batch in a single class</param>
		/// <param name="batchSize">Size of a single batch. Depends on <paramref name=synchronizedUpdate""/> parameter.
		/// For synchronized update means the total number of documents in all classes.
		/// For unsynchronized update means the number of documents in a single class.</param>
		/// <param name="synchronizedUpdate">Decides whether all classes are updated synchronously or not.</param>
		public StreamClassifier(double minsup, int batchSize, bool synchronizedUpdate, bool isWindow, int windowSize, bool weightSize, bool weightSupport)
		{
			Pattern.WeightSize = weightSize;
			Pattern.WeightSupport = weightSupport;

			this.models = new List<ClassModel>();
			this.documentBuffer = new List<Document>();

			this.minsup = minsup;
			this.treeMiningBatchSize = batchSize;
			this.synchronizedUpdate = synchronizedUpdate;

			this.isWindow = isWindow;
			this.windowSize = windowSize;

			this.totalNumberOfDocuments = 0;
			this.totalNumberOfCorrectlyAssignedDocuments = 0;
			this.numberOfCorrectlyAssignedDocumentsInLastBatch = 0;
			this.numberOfDocumentsInSamplingBatch = 0;
		}

		/// <summary>
		/// Updates the classifier with a new document.
		/// Adds new document to a specific class model and, depending on a chosen update model, updates the model if a maximal batch size is reached.
		/// </summary>
		/// <param name="document">New document</param>
		public void UpdateModel(Document document)
		{
			if (!models.Any(x => x.Class == document.ExpectedClass))
			{
				this.models.Add(new ClassModel(document.ExpectedClass, this.minsup, this.treeMiningBatchSize, this.isWindow, this.windowSize));
			}
			var model = this.models.Single(x => x.Class == document.ExpectedClass);
			model.AddDocument(document);

			if (this.synchronizedUpdate)
			{
				if (this.totalNumberOfDocuments % this.treeMiningBatchSize == 0)
				{
					foreach (var m in this.models)
					{
						m.UpdateModel(this.synchronizedUpdate);
					}
				}
			}
			else
			{
				model.UpdateModel(this.synchronizedUpdate);
			}
		}

		/// <summary>
		/// Classifies a document according to current model.
		/// </summary>
		/// <param name="document">The document</param>
		/// <param name="range">Establishes the boundries of a neighbourhood for the Nearest Patterns algorithm.</param>
		public void ClassifyDocument(Document document)
		{
			if (models.Any(x => x.Model.Any()))
			{
				Classifier.Classify(this.GetCompleteModel(), document);
				if (document.ExpectedClass == document.ActualClass)
				{
					this.totalNumberOfCorrectlyAssignedDocuments++;
					this.numberOfCorrectlyAssignedDocumentsInLastBatch++;
				}
				this.totalNumberOfDocuments++;
				this.numberOfDocumentsInSamplingBatch++;
			}

			this.documentBuffer.Add(document);
		}

		/// <summary>
		/// Clears the statistics used for evaluation of the current batch.
		/// </summary>
		public void ClearSamplingStats()
		{
			this.numberOfCorrectlyAssignedDocumentsInLastBatch = 0;
			this.numberOfDocumentsInSamplingBatch = 0;
		}

		/// <summary>
		/// Returns all patterns form all models in a single collection
		/// </summary>
		/// <returns>Combined model for all classes</returns>
		private IEnumerable<Pattern> GetCompleteModel()
		{
			foreach (var model in models)
			{
				foreach (var pattern in model.Model)
				{
					yield return pattern;
				}
			}
		}
	}
}
